# -*- coding: utf-8 -*-

from numpy import zeros, zeros_like, dot, mean
from numpy.random import permutation
from contextlib import contextmanager

class BaseClassifier(object):

    """
    The base class of a linear classifier. This class shouldn't be explicitly
    used but instead child classes (e.g., missmatch.Perceptron) should be used.
    Notice that the `repr` of a BaseClassifier is the array of its weights.

    Attributes
    ----------
    w : array
        output weights
    """

    def __init__(self):

        self.w = None # TODO: should turn this into a property
        self._init()

    def train(self, x, y, n_epochs, verbose=False,
              perms=None, monitors=None, reset=True):
        """
        b.train(x, y, 10, monitors=[0])

        updates the weights of the classifier repeatedly for n_epochs epochs

        Parameters
        ----------
        x, y : arrays
            patterns and labels of the training set
        n_epochs : int
            number of epochs
        verbose : bool
            if True useful information is printed during learning
        perms : list of arrays or None
            list of arrays of indices of patterns to be presented on each epoch
        monitors : array or None
            list of indices of weights to be monitored (WARNING: unused atm!)
        reset : bool
            whether to reset the weight matrix to zeros prior to training
        """

        # TODO: monitors should be external classes bound to the classifier

        p = x.shape[1]
        N = x.shape[0]

        if reset:
            self.reset(N)

        #traces = []
        if perms is None:
            perms = [permutation(p) for i in xrange(n_epochs)]

        for n in xrange(n_epochs):
            with self._train_context():
                for i in xrange(p):
                    if verbose:
                        print "-- pattern", perms[n][i], "label", y[perms[n][i]]
                    self._update(x[:, perms[n][i]], y[perms[n][i]])
                    #if monitors is not None:
                        #traces.append(self.w[monitors])
        return

    def predict(self, x):
        '''
        output = b.predict(x)

        produces an output given an input pattern using the internal parameters

        Parameters
        ----------

        x : array
            input pattern

        Returns
        -------
        out : float
            the activation level
        '''
        with self._predict_context():
            return self._predict(x)

    def test(self, x, y):
        """
        perf = b.test(x, y) * 100

        returns the hit rate on a given dataset

        Parameters
        ----------
        x, y : arrays
            input patterns and labels

        Returns
        -------
        rate : floate
            the hit rate
        """
        with self._test_context():
            return mean((self._predict(x) * y) > 0)

    def update(self, x, y):
        """
        b.update(x, y)

        given an input pattern and an output label, it updates internal
        parameters such as synaptic weights

        Parameters
        ----------
        x : array
            input pattern
        y : label
            pattern label (+1 or -1)

        Returns
        -------
        None
        """
        with self._update_context():
            return self._update(x, y)

    def reset(self, N): # this is ugly because of N
        """
        b.reset(N)

        reset the internal parameters to initial values, e.g., synaptic weights
        to 0

        Parameters
        ----------
        N : int
            size of the classifier

        Returns
        -------
        None
        """
        self.w = zeros(N)
        with self._reset_context():
            self._reset()

    def _predict(self, x):
        """
        This function returns the response of the classifier and is wrapped
        within the `__call__` method. The default implementation returns a dot
        product between the input and the weights.
        """
        return dot(self.w, x)

    def _init(self):
        """
        This function is offered to implement necessary operations required for
        initializing the classifier. The default implementation does nothing.
        See also `reset()`.
        """
        pass

    @contextmanager
    def _predict_context(self):
        try:
            self._predict_setup()
            yield
        finally:
            self._predict_teardown()


    def _predict_setup(self):
        """
        This function is used in the `predict` method. It is called before
        `_predict` and is offered to implement all the necessary operations to
        correctly operate the `predict` function. The default implementation
        does nothing.
        """
        pass

    def _predict_teardown(self):
        """
        This function is used in the `predict` method. It is called after
        `_predict` and is offered to implement all the necessary operations to
        correctly operate the `predict` function. The default implementation
        does nothing.
        """
        pass

    @contextmanager
    def _train_context(self):
        try:
            self._train_setup()
            yield
        finally:
            self._train_teardown()

    def _train_setup(self):
        """
        This function is used in the `train` method. It is called before
        `_train` and is offered to implement all the necessary operations to
        correctly operate the `train` function. The default implementation
        does nothing.
        """
        pass

    def _train_teardown(self):
        """
        This function is used in the `train` method. It is called after
        `_train` and is offered to implement all the necessary operations to
        correctly operate the `train` function. The default implementation
        does nothing.
        """
        pass

    @contextmanager
    def _update_context(self):
        try:
            self._update_setup()
            yield
        finally:
            self._update_teardown()

    def _update_setup(self):
        """
        This function is used in the `update` method. It is called before
        `_update` and is offered to implement all the necessary operations to
        correctly operate the `update` function. The default implementation
        does nothing.
        """
        pass

    def _update_teardown(self):
        """
        This function is used in the `update` method. It is called after
        `_update` and is offered to implement all the necessary operations to
        correctly operate the `update` function. The default implementation
        does nothing.
        """
        pass

    @contextmanager
    def _test_context(self):
        try:
            self._test_setup()
            yield
        finally:
            self._test_teardown()

    def _test_setup(self):
        """
        This function is used in the `test` method. It is called before
        `_test` and is offered to implement all the necessary operations to
        correctly operate the `test` function. The default implementation
        does nothing.
        """
        pass

    def _test_teardown(self):
        """
        This function is used in the `test` method. It is called after
        `_test` and is offered to implement all the necessary operations to
        correctly operate the `test` function. The default implementation
        does nothing.
        """
        pass

    @contextmanager
    def _reset_context(self):
        try:
            self._reset_setup()
            yield
        finally:
            self._reset_teardown()

    def _reset_setup(self):
        """
        This function is used in the `reset` method. It is called before
        `_reset` and is offered to implement all the necessary operations to
        correctly operate the `reset` function. The default implementation
        does nothing.
        """
        pass

    def _reset_teardown(self):
        """
        This function is used in the `reset` method. It is called after
        `_reset` and is offered to implement all the necessary operations to
        correctly operate the `reset` function. The default implementation
        does nothing.
        """
        pass

    def _update(self, x, y):
        """
        This function is called at every train cycle and is supposed to update
        the internal weights according to a synaptic update rule. The default
        implementation does nothing.
        """
        pass

    def _reset(self):
        """
        This function is offered to implement necessary operations to reset the
        status of the classifier to its initial conditions. The default
        implementation does nothing.
        """
        pass

    def __call__(self, x):
        """x.__call__(...) <==> x(...)"""
        return self.predict(x)

    def __getitem__(self, i):
        """x.__getitem__(i) <==> x[i]"""
        return self.w[i]

    def __getslice__(self, i, j):
        """x.__getslice__(i, j) <==> x[i:j]"""
        return self.w[i:j]

    def __len__(self):
        """x.__len__() <==> len(x)"""
        return len(self.w)

    def __repr__(self):
        return str(self.w)
