# -*- coding: utf-8 -*-
# @author Mattia Rigotti
# @date Fri Jun 13 16:18:51 EDT 2014
from __future__ import division
import os
import pandas as pd
import numpy as np


DATA_DIR = '../data'

# File extensions that will be recognized as databases
EXT_DATABASE = ['csv']


def seek_data(directory=DATA_DIR):
    """
    dataset = seek_data(directory=DATA_DIR)

    Returns a list of the datasets in directory

    Parameters
    ----------
    directory : string, optional
        directory that needs to be inspected for datasets

    Returns
    -------
    dataset : tuple
        tuple that lists the file name of available datasets
    """
    dirlist = []
    try:
        dirlist = os.listdir(directory)
    except OSError:
        raise Exception("I cannot find directory " + directory)
    finally:
        dataset = []
        for filename in dirlist:
            if filename[-3:] in EXT_DATABASE:
                dataset.append(directory + '/' + filename)
        return tuple(dataset)


def load_data(filename):
    """
    inputs, labels = load_data(dataset[0])

    Parameters
    ----------
    filename : string
        file name of dataset

    Returns
    -------
    inputs, labels : numpy arrays
    """
    inputs = pd.read_csv(filename, header=None)
    labels = pd.Categorical.from_array(inputs[inputs.shape[1] - 1])
    inputs = inputs.drop(inputs.shape[1] - 1, axis=1)
    return np.array(inputs).T, np.array(labels)


def describe_data(filename):
    """
    dataset_desc = describe_data(dataset[0])

    Parameters
    ----------
    filename : string
        file name of dataset

    Returns
    -------
    dataset_sec : dictionary
        description of dataset (name, size, classes, etc.)
    """
    inputs, labels = load_data(filename)

    dataset_desc = {'name': os.path.basename(filename),
                    'dir': os.path.dirname(filename),
                    'samples': inputs.shape[1],
                    'features': inputs.shape[0],
                    'classes': len(np.unique(labels)),
                    'miss_data': sum(np.isnan(inputs.flatten()))
                    }
    return dataset_desc


def sample_train_test(inputs, labels, train_fraction=0.5, stratified=True,
                      normalization='zscore', cutoff=0.05, nan_to_num=True):
    """
    xi, yi, xt, yt = sample_train_test(inputs, labels, train_fraction=0.9,
                                       stratified=True, cutoff=0.1)

    Devides dataset in training and test set

    Parameters
    ----------
    inputs : array
        input patterns

    labels : array
        labels of patterns

    train_fraction : double, optional
        fraction of patterns assigned to training set

    stratified : bool, optional
        whether sampling is stratified within classes

    normalization : string, optional
        what type of normalization should be applied to the data
        (training set is used to determine normalization factors)
        possible values: 'zscore', 'zca', None

    cutoff : double, optional
        cutoff to be applied to zscore normalization

    nan_to_num : bool, optional
        whether empty features have to be replaced with 0

    Returns
    -------
    xi : array
        training set of patterns

    yi : array
        labels of training set

    xt : array
        test set of patterns

    yt : array
        labels of test set
    """
    # Determine training and test set
    ind_train = []
    ind_test = []

    if stratified:
        labs = np.unique(labels)
        for l in labs:
            ind = np.nonzero((labels == l).flatten())[0]
            np.random.shuffle(ind)
            ind_train = ind_train + \
                ind[:round(train_fraction * len(ind))].tolist()
            ind_test = ind_test + \
                ind[round(train_fraction * len(ind)):].tolist()
    else:
        ind = np.linspace(0, len(labels) - 1, len(labels))
        np.random.shuffle(ind)
        ind_train = ind_train + ind[:round(train_fraction * len(ind))].tolist()
        ind_test = ind_test + ind[round(train_fraction * len(ind)):].tolist()
    
    # Normalize
    xi = inputs[:, ind_train]
    yi = labels[ind_train]
    xt = inputs[:, ind_test]
    yt = labels[ind_test]

    if normalization is None:
        pass

    elif normalization == 'zscore':
        m = np.nanmean(xi, axis=1)
        s = np.nanstd(xi, axis=1) + cutoff
        xi = (xi - m[:, np.newaxis]) / (s[:, np.newaxis] + cutoff)
        xt = (xt - m[:, np.newaxis]) / (s[:, np.newaxis] + cutoff)

    elif normalization == 'zca':
        m = np.nanmean(xi, axis=1)
        S = np.cov(np.nan_to_num(xi))
        lam, vec = np.linalg.eigh(S)    # eigh is for symmetric matrices
        ZCA = np.dot(vec / (np.sqrt(lam) + cutoff), vec.T)
        xi = np.dot(ZCA, xi - m[:, np.newaxis])
        xt = np.dot(ZCA, xt - m[:, np.newaxis])

    else:
        raise Exception("Normalization type \"" + str(normalization) +
                        "\" not defined")
    if nan_to_num:
        return np.nan_to_num(xi), yi, np.nan_to_num(xt), yt
    else:
        print "END"
        return xi, yi, xt, yt
