# -*- coding: utf-8 -*-

import numpy as np

from numpy.random import permutation


class Ensemble(object):

    """
    The Ensemble class encapsulates multiple classifiers into a single object.

    Attributes
    ----------
    ECOC : array
        the ECOC matrix
    labels : array
        after learning, it contains the list of known labels
    w : list of classifier objects
        after learning, the list of classifier objects

    Example
    -------
    import os
    import missmatch as mm

    DATA_DIR = os.path.dirname(mm.__file__) + '/../tests/data'

    datasets = mm.seek_data(DATA_DIR)
    inputs, labels = dm.load_data(datasets[0])
    xi, yi, xt, yt = dm.sample_train_test(inputs, labels)
    e = mm.Ensemble()
    e.train(xi, yi, n_epochs=10, unit_type=mm.Perceptron, code_args={'n':10})
    perf = e.test(xt, yt)
    print "Performance is", perf * 100
    """

    def __init__(self):
        self.reset()

    def reset(self):
        """
        e.reset()

        Reset the internal states of the Ensemble.
        """
        self.ECOC = None
        self.labels = None
        self.w = None

    def train(self, x, y, n_epochs, unit_type, verbose=False,
              code='one-vs-rest', code_args={},
              balance='max', reset=True,
              **kwargs):
        """
        e.train(x, y, 50, missmatch.Perceptron, code_args={'n':10}, lrate=0.01)

        Parameters
        ----------
        x : array
            input patterns

        y : numpy array
            pattern labels

        n_epochs : int
            number of epochs

        unit_type : class
            the class of classifiers

        verbose : bool
            if True useful information is printed during training

        code : array or string, optional
            type of output code to turn multi-class problem to sequence of
            binary ones (possible values: 'one-vs-rest', 'pairs', or a generic
            ECOC matrix)

        code_args : dict
            dictionary of parameters for given code (e.g., if code is
            'one-vs-rest' then {'n':10} will create an ECOC matrix of 10 units
            per label)

        balance : string or integer, optional
            whether positive and negative classes should be balanced and how
            (possible values: 'max', sample from as many patterns as in the
            class with most patterns; 'min', analogous as before; if it is an
            integer sample as many patterns in positive and negative class as
            the given value)

        reset : bool
            whether to reset the weight matrix to zeros prior to training

        **kwargs : keyword arguments
            passed to classifiers' init functions

        Returns
        -------
        None
        """
        self.labels = np.sort(np.unique(y))
        self._set_ECOC(code, code_args)

        if reset:
            self.w = [unit_type(**kwargs) for n in xrange(self.ECOC.shape[0])]

        for n in xrange(self.ECOC.shape[0]):
            if verbose:
                print "- Training unit", n
                print "    assigned to labels:"
                print "    ", self.labels[self.ECOC[n] > 0]
                print "    and not to labels:"
                print "    ", self.labels[self.ECOC[n] < 0]
                print "    disregarding labels:"
                print "    ", self.labels[self.ECOC[n] == 0]
            ind_pos, ind_neg = self._which_inds(n, y)

            # balancing
            if balance == 'max':
                n_samp = np.max((len(ind_pos), len(ind_neg)))

            elif balance == 'min':
                n_samp = np.min((len(ind_pos), len(ind_neg)))

            elif type(balance) == int:
                n_samp = balance

            else:
                raise Exception("Argument balance has a non-supported type")

            ind_all = int(n_samp / len(ind_pos)) * ind_pos +\
                ind_pos[:np.mod(n_samp, len(ind_pos))]
            ind_all += int(n_samp / len(ind_neg)) * ind_neg +\
                ind_neg[:np.mod(n_samp, len(ind_neg))]
            np.random.shuffle(ind_all)

            y_inds = np.r_[[np.argwhere(l == self.labels)
                            for l in y[ind_all]]].flatten()
            self.w[n].train(x[:, ind_all], self.ECOC[n, y_inds], n_epochs,
                            reset=reset, verbose=verbose)

    def activate(self, x):
        """
        e.activate(self, x)

        returns the activation values of all the units

        Parameters
        ----------
        x : array
            input pattern (size must match internal weight matrix after learning)

        Returns
        -------
        out : array
            activation values
        """
        H = np.r_[[w.predict(x) for w in self.w]]
        return H

    def predict(self, x):
        """
        e.predict(self, x)

        returns the predicted label for the given input pattern

        Parameters
        ----------
        x : array
            input pattern (size must match internal weight matrix after learning)

        Returns
        -------
        pred : same type as the elements of e.labels
            the predicted label
        """
        H = np.dot(self.ECOC.T, self.activate(x))
        return self.labels[np.argmax(H, axis=0)]

    def test(self, x, y):
        """
        e.test(x, y)

        test ensemble's performance on a dataset

        Parameters
        ----------
        x, y : arrays
            input patterns and output labels

        Returns
        -------
        performance : float
            hit rate
        """
        return np.mean(self.predict(x) == y)

    def confusion(self, x, y):
        """
        e.confusion(x, y)

        returns the confusion matrix on the given dataset

        Parameters
        ----------
        x, y : arrays
            input patterns and labels

        Returns
        -------
        conf : array
            confusion matrix
        """

        outs = self.predict(x)
        conf = np.r_[[[sum((outs == l) * (y == m))\
                       for l in self.labels]\
                      for m in self.labels]]

        return conf

    def pe(self, x, y):
        """
        e.pe(x, y)

        Ths function returns the upper bound of the generalization error à la
        Breiman [1]:
            $$
            PE = \leq \bar \rho (1-s^2)s^2
            $$
        where $\bar \rho$ is the mean value of the correlation of the
        classifiers' responses and $s$ is the average 'strength' of the
        classifiers.

        [1] Breiman, Leo. "Random forests." Machine learning 45.1 (2001): 5-32.

        Parameters
        ----------
        x, y : arrays
            input patterns and labels

        Returns
        -------
        pe : float
            the generalization error
        """
        strength = self.strength(x, y)
        pe = self.correlation(x) * (1 - strength ** 2) / strength ** 2
        return pe

    def strength(self, x, y):
        """ens.strength(x, y)<==> 1 - ens.errors(x, y)"""
        return 1 - self.error(x, y)

    def error(self, x, y):
        """
        e.error(x, y)

        computes the average error of single classifiers on a given dataset

        Parameters
        ----------
        x, y : arrays
            input patterns and labels

        Returns
        -------
        err : float
            average error
        """
        normal_ = 1./self.ECOC.shape[0]
        err = 0.
        for n in xrange(self.ECOC.shape[0]):
            ind_pos, ind_neg = self._which_inds(n, y)
            ind_all = np.concatenate((ind_pos, ind_neg))
            y_inds = np.r_[[np.argwhere(l == self.labels)
                            for l in y[ind_all]]].flatten()
            y_code = self.ECOC[n, y_inds]
            err_tmp = 1 - np.mean((self.w[n](x[:, ind_all]) * y_code) > 0)
            err += err_tmp * normal_
        return err

    def correlation(self, x):
        """
        e.correlation(x)

        computes the correlation of output nodes on a given dataset

        Parameters
        ----------
        x : arrays
            input patterns

        Returns
        -------
        corr : float
            correlation of responses of output units
        """
        responses = self.activate(x) > 0
        # important: use np.sum
        corr = 1. * np.sum(np.cov(responses) * (1. - np.eye(self.ECOC.shape[0])))\
                / self.ECOC.shape[0] / (self.ECOC.shape[0] - 1)
        return corr

    def _which_inds(self, n, y):
        """
        e._which_inds(3, y)

        returns all the indices of positive and negative elements corresponding
        to a given dataset labels list for a given unit of the ensemble

        Parameters
        ----------
        n : int
            index of the unit

        y : array
            labels of the dataset

        Returns
        -------
        ind_pos, ind_neg : arrays
            arrays of indices of positive and negative elements
        """
        lab_pos = self.labels[np.nonzero(self.ECOC[n, :] == 1)[0]]
        lab_neg = self.labels[np.nonzero(self.ECOC[n, :] == -1)[0]]
        ind_pos = [i[0] for i in enumerate(y) if i[1] in lab_pos]
        ind_neg = [i[0] for i in enumerate(y) if i[1] in lab_neg]
        return ind_pos, ind_neg

    def _set_ECOC(self, code, code_args):
        """
        e._set_ECOC('one-vs-rest', 4, 20)

        create the ECOC matrix of +/- ones according to given code, for the
        given number of labels and given number of units

        Parameters
        ----------
        code : array or string
            type of output code to tun multi-class problem into a sequence of
            binary ones (possible values are 'one-vs-rest', 'pairs', 'n-vs-rest'
            or a generic ECOC matrix)

        n_labels : int
            number of labels (determines the number of columns)

        code_args : dict
            dictionary of parameters for given code (e.g., if code is
            'one-vs-rest' then {'n':10} will create an ECOC matrix of 10 units
            per label)

        Returns
        -------
        None
        """
        if code == 'one-vs-rest':
            n_labels = len(self.labels)
            ECOC = -np.ones((n_labels, n_labels)) + 2 * np.eye(n_labels)
            # parse code parameters
            if 'n' in code_args:
                ECOC = np.row_stack([ECOC] * code_args['n'])

        elif code == 'pairs':
            n_labels = len(self.labels)
            ECOC = np.zeros((n_labels * (n_labels - 1) / 2, n_labels))
            t = 0
            for i in xrange(n_labels):
                for j in xrange(i + 1, n_labels):
                    ECOC[t, i] = 1
                    ECOC[t, j] = -1
                    t += 1
            # parse code parameters
            if 'n' in code_args:
                ECOC = np.row_stack([ECOC] * code_args['n'])

        elif code == 'n-vs-rest':
            try:
                n_hot = code_args['n_hot']
            except KeyError:
                n_hot = 1
            try:
                b = code_args['b']
                n = 2 ** b
            except KeyError:
                b = n_hot
                n = 0
            n_labels = len(self.labels)
            # count how many bits are needed for n-hot code
            f = np.math.factorial
            while n < n_labels:
                b += 1
                n = f(b) / f(n_hot) / f(b - n_hot)
            ECOC = -np.ones((b, n_labels))
            # pick up the binary words with n_hot ones
            ecoc_columns = [m[2:] for m in map(bin, range(2 ** b))\
                            if sum([int(k) for k in m[2:]]) == n_hot]
            # substitute -1 with 1 in the final ECOC according to choosen words
            for i, l in enumerate(permutation(ecoc_columns)[:n_labels]):
                for j, b in enumerate(l[::-1]):
                    ECOC[:, i][j] = (1 if b == '1' else -1)
            # parse other code parameters
            if 'n' in code_args:
                ECOC = np.row_stack([ECOC] * code_args['n'])

        elif isinstance(code, type(np.array([]))):
            if code.shape[1] == len(self.labels):
                ECOC = code
            else:
                raise Exception(
                    "Matrix code must have one column per label.")
        else:
            raise Exception("Argument code has a non-supported format.")

        self.ECOC = ECOC

    def __len__(self):
        """x.__len__() <==> len(x)"""
        return len(self.w)

    def __getslice__(self, i, j):
        """x.__getslice__(i, j) <==> x[i:j]"""
        return self.w[i:j]

    def __getitem__(self, i):
        """x.__getitem__(i) <==> x[i]"""
        return self.w[i]

    def __call__(self, y):
        """x.__call__(y) <==> x(y) """
        return self.predict(y)
