# -*- coding: utf-8 -*-

from numpy import dot, sqrt, mean, size, zeros_like, where
from numpy.random import permutation, rand, normal

from base_classifier import BaseClassifier

# TODO: fix arguments passing for Ensemble


class Perceptron(BaseClassifier):

    """
    Rosenblatt's perceptron.

    Parameters
    ----------
    lrate : float, optional
        learning rate
    threshold : float, optional
        output threshold
    delta : float, optional
        margin

    Attributes
    ----------
    lrate : float
        learning rate
    threshold : float
        output threshold
    delta : float
        margin
    w : array
        output weights

    Example
    -------
    import os
    import missmatch as mm

    DATA_DIR = os.path.dirname(mm.__file__) + '/../tests/data'

    datasets = mm.seek_data(DATA_DIR)
    inputs, labels = mm.load_data(datasets[0])

    xi, yi, xt, yt = mm.sample_train_test(inputs, labels)

    p = mm.Perceptron(lrate=0.05)
    p.train(xi, yi)
    perf = p.test(xt, yt)
    print "Performance is", perf * 100
    """

    def __init__(self, lrate=0.01, threshold=0, delta=0):
        """
        p = missmatch.Perceptron(lrate=0.02)
        """

        super(Perceptron, self).__init__()

        self.lrate = lrate
        self.thr = threshold
        self.delta = delta

    def _update(self, x, y):
        h = self(x)
        dh = y * h
        if dh <= self.delta * sqrt(dot(self.w, self.w)):
            self.w += self.lrate * y * x
        return


class SoftBounds(BaseClassifier):

    """
    A perceptron with soft bounds, i.e., synapses that exponentially approach
    hard bounds.

    Parameters
    ----------
    lrate : float, optional
        learning rate
    threshold : float, optional
        output threshold
    delta : float, optional
        margin
    wmax, wmin : float, optional
        maximum and minimum weight values

    Attributes
    ----------
    lrate : float
        learning rate
    threshold : float
        output threshold
    delta : float
        margin
    wmax, wmin : float
        maximum and minimum weight values
    w : array
        output weights

    Example
    -------
    import os
    import missmatch as mm

    DATA_DIR = os.path.dirname(mm.__file__) + '/../tests/data'

    datasets = mm.seek_data(DATA_DIR)
    inputs, labels = mm.load_data(datasets[0])

    xi, yi, xt, yt = mm.sample_train_test(inputs, labels)

    p = mm.SoftBounds(lrate=0.05)
    p.train(xi, yi)
    perf = p.test(xt, yt)
    print "Performance is", perf * 100
    """

    def __init__(self, lrate=0.01, threshold=0, delta=0, wmax=1., wmin=0.):

        super(SoftBounds, self).__init__()

        self.lrate = lrate
        self.thr = threshold
        self.delta = delta
        self.wmax = wmax
        self.wmin = wmin

    def _update(self, x, y):
        h = self(x)
        dh = y * h
        if dh <= self.delta * sqrt(dot(self.w, self.w)):
            delta_w = self.lrate * y * x
            self.w += delta_w * ((delta_w > 0) * (self.wmax - self.w) -
                                 (delta_w < 0) * (self.wmin + self.w))


class BinaryStoch(BaseClassifier):

    """
    A perceptron with binary, stochastic synapses.

    Parameters
    ----------
    threshold : float, optional
        output threshold
    delta : float, optional
        margin
    pltp, pltd : float, optional
        probabilities of LTP and LTD
    g_i : float, optional
        global inhibition term

    Attributes
    ----------
    threshold : float
        output threshold
    delta : float
        margin
    pltp, pltd : float
        probabilities of LTP and LTD
    g_i : float
        global inhibition term
    w : array
        output weights

    Example
    -------
    import os
    import missmatch as mm

    DATA_DIR = os.path.dirname(mm.__file__) + '/../tests/data'

    datasets = mm.seek_data(DATA_DIR)
    inputs, labels = mm.load_data(datasets[0])

    xi, yi, xt, yt = mm.sample_train_test(inputs, labels)

    p = mm.BinaryStoch(pltp=0.05, pltd=0.05)
    p.train(xi, yi)
    perf = p.test(xt, yt)
    print "Performance is", perf * 100
    """

    def __init__(self, threshold=0, delta=0,
                 pltp=0.01, pltd=0.01, g_i=0.5):

        super(BinaryStoch, self).__init__()

        self.thr = threshold
        self.delta = delta
        self.pltp = pltp
        self.pltd = pltd
        self.g_i = g_i

    def _update(self, x, y):
        h = self._predict(x)
        dh = y * h
        if y < 0 and h >= -self.delta * sqrt(dot(self.w, self.w)):
            to_switch = rand(size(self.w)) < self.pltd
            self.w -= (x > 0) * to_switch * self.w
        elif y > 0 and h <= self.delta * sqrt(dot(self.w, self.w)):
            to_switch = rand(size(self.w)) < self.pltp
            self.w += (x > 0) * to_switch * (1 - self.w)

    def _predict(self, x):
        return dot(self.w - self.g_i, x)


class BinaryStochMismatch(BaseClassifier):

    """
    A perceptron with binary, stochastic synapses and parameter variations.

    Parameters
    ----------
    threshold : float, optional
        output threshold
    delta : float, optional
        margin
    pltp, pltd : float, optional
        probabilities of LTP and LTD
    g_i : float, optional
        global inhibition term
    only_positive : bool
        whether to clip all negative weights to 0
    w_hi_mismatch : float
        variance of the gaussian distribution of weights values

    Attributes
    ----------
    threshold : float
        output threshold
    delta : float
        margin
    pltp, pltd : float
        probabilities of LTP and LTD
    g_i : float
        global inhibition term
    w : array
        output weights (w = w_s * w_hi)
    w_s : array
        synaptic states (0/1)
    w_hi : array
        synaptic weights of active synapses

    Example
    -------
    import os
    import missmatch as mm

    DATA_DIR = os.path.dirname(mm.__file__) + '/../tests/data'

    datasets = mm.seek_data(DATA_DIR)
    inputs, labels = mm.load_data(datasets[0])

    xi, yi, xt, yt = dm.sample_train_test(inputs, labels)

    p = mm.BinaryStochMismatch(pltp=0.05, pltd=0.05, w_hi_mismatch=0.5)
    p.train(xi, yi)
    perf = p.test(xt, yt)
    print "Performance is", perf * 100

    See also
    --------
    BinaryStoch

    """

    def __init__(self, threshold=0, delta=0,
                 pltp=0.01, pltd=0.01,
                 w_hi_mismatch=0, only_positive=True,
                 g_i=0.5):

        super(BinaryStochMismatch, self).__init__()

        self.thr = threshold
        self.delta = delta
        self.pltp = pltp
        self.pltd = pltd
        self.g_i = g_i
        self.w_hi = 1. - (2. * normal(0, w_hi_mismatch, size(self.w)) - 1)\
                if w_hi_mismatch > 0 else 1.
        if only_positive:
            self.w_hi = where(self.w_hi > 0, self.w_hi, 0)

    def _update(self, x, y):
        h = self._predict(x)
        dh = y * h
        if y < 0 and h >= -self.delta * sqrt(dot(self.w, self.w)):
            to_switch = rand(size(self.w)) < self.pltd
            self.w_s -= (x > 0) * to_switch * self.w_s
        elif y > 0 and h <= self.delta * sqrt(dot(self.w, self.w)):
            to_switch = rand(size(self.w)) < self.pltp
            self.w_s += (x > 0) * to_switch * (1 - self.w_s)
        self.w = self._states_to_weights()

    def _predict(self, x):
        self.w = self._states_to_weights()
        return dot(self.w - self.g_i, x)

    def _states_to_weights(self):
        return self.w_hi * self.w_s

    def _reset(self):
        self.w_s = zeros_like(self.w)


class BinaryStochNoisy(BaseClassifier):

    """
    A perceptron with binary, stochastic synapses and noisy probabilities of LTP
    and LTD. A noise contribution is added at each pattern presentation.

    Parameters
    ----------
    threshold : float, optional
        output threshold
    delta : float, optional
        margin
    pltp, pltd : float, optional
        probabilities of LTP and LTD
    g_i : float, optional
        global inhibition term
    noise_level : float, optional
        noise level applied to LTP and LTD probabilities, applied at each
        pattern presentation

    Attributes
    ----------
    threshold : float
        output threshold
    delta : float
        margin
    pltp, pltd : float
        probabilities of LTP and LTD
    g_i : float
        global inhibition term
    w : array
        output weights

    Example
    -------
    import os
    import missmatch as mm

    DATA_DIR = os.path.dirname(mm.__file__) + '/../tests/data'

    datasets = mm.seek_data(DATA_DIR)
    inputs, labels = mm.load_data(datasets[0])

    xi, yi, xt, yt = mm.sample_train_test(inputs, labels)

    p = mm.BinaryStochNoisy(pltp=0.05, pltd=0.05)
    p.train(xi, yi)
    perf = p.test(xt, yt)
    print "Performance is", perf * 100
    """

    def __init__(self, threshold=0, delta=0,
                 pltp=0.01, pltd=0.01, g_i=0.5,
                 noise_level=0.01):

        super(BinaryStoch, self).__init__()

        self.thr = threshold
        self.delta = delta
        self.pltp = pltp
        self.pltd = pltd
        self.g_i = g_i

    def _update(self, x, y):
        h = self._predict(x)
        dh = y * h
        pltpd_noise = normal(0, noise_level)
        if y < 0 and h >= -self.delta * sqrt(dot(self.w, self.w)):
            to_switch = rand(size(self.w)) < (self.pltd + pltpd_noise)
            self.w -= (x > 0) * to_switch * self.w
        elif y > 0 and h <= self.delta * sqrt(dot(self.w, self.w)):
            to_switch = rand(size(self.w)) < (self.pltp + pltpd_noise)
            self.w += (x > 0) * to_switch * (1 - self.w)

    def _predict(self, x):
        return dot(self.w - self.g_i, x)

